// quickselect.cpp
// VERSION 2
// Glenn G. Chappell
// 2 Oct 2013
//
// For CS 411/611 Fall 2013
// Lomuto Partition & Quickselect

#include <iostream>
using std::cout;
using std::endl;
using std::cin;
#include <algorithm>
using std::iter_swap;
using std::swap;
#include <vector>
using std::vector;
#include <cstdlib>
using std::size_t;
#include <cassert>
// For assert
#include <cmath>
using std::sqrt;


// lpartition
// Partitions a sequence about a given pivot. Uses Lomuto Algorithm.
// Returns the new pivot position via a reference parameter.
// 
// Requirements on Types:
//     FDIter must be a forward iterator type.
//     operator< must be a total order on the value type of FDIter.
// Pre:
//     [first, last) is a valid nonempty range.
//     pivotiter is an iterator in the range [first, last).
// Post:
//     [first, last) contains the same items as it did initially,
//      but now each ITEM before pivotiter has ITEM < *pivotiter,
//      and each ITEM after pivotiter has !(ITEM < *pivotiter).
//     pivotiter lies in [first, last) and references an item with
//      the same value as *pivotiter before the function call.
template <typename FDIter>
void lpartition(FDIter first, FDIter last,  // Range to partition
                FDIter & pivotiter)         // Iterator to pivot
{
    // Put the pivot at the start of the list
    iter_swap(first, pivotiter);

    // Now "first" points to the pivot

    // Make the "left list": list of items less than pivot
    FDIter leftFinal = first;  // points to final item in left list
    FDIter check = first;      // item to check
    for (++check;              // start after pivot, iterate through list
         check != last;
         ++check)
    {
        if (*check < *first)              // if item < PIVOT
        {
            ++leftFinal;                  // move up end mark
            iter_swap(leftFinal, check);  // and put item in left list
        }
    }

    // Note new pivot position, for caller, and put pivot there
    pivotiter = leftFinal;
    iter_swap(first, pivotiter);
}


// medianOf3
// Given 3 iterators, finds the median of the values they reference and
//  returns an iterator to it. Values are not altered; return value is
//  always equal to one of the parameters.
// Performs at most 3 comparisons.
// Stable (in the non-mutating selection-algorithm sense).
template <typename Iter>
Iter medianOf3(Iter ai, Iter bi, Iter ci)
{
    auto & a(*ai);
    auto & b(*bi);
    auto & c(*ci);

    return b < a ? c < a ? c < b ? bi : ci : ai
                 : c < b ? c < a ? ai : ci : bi;
}


// quickselect
// Selection using Quickselect algorithm. [first, last) specifies a
//  nonempty range. index is an index in this range (0 .. size-1).
//  Return value is an iterator to the item in this range that would be
//  at the given index, if the range were sorted (not necessarily in a
//  stable manner). The range is modified, but not sorted.
// Uses lpartition, medianOf3
// Pre:
//     [first, last) is a nonempty valid range.
//     0 <= index < (last-first).
// Post:
//     Return is iterator to the item that would be at position given by
//      index, if the data were sorted (not necessarily in a stable
//      manner).
//     Data in [first, last) are same items as initially, but possibly
//      in a different order.
template <typename RAIter>
RAIter quickselect(RAIter first, RAIter last,  // Range: [first, last)
                   size_t index)               // Selection index
{
    while (true)  // For tail-recursion elimination
    {    
        // Find median-of-three pivot & point pivotiter at it
        size_t size = last - first;
        assert (size != 0);
        RAIter pivotiter = medianOf3(first, first+size/2, last-1);
    
        // Do partition
        lpartition(first, last, pivotiter);
        size_t pivotindex = pivotiter - first;
    
        // BASE CASE of former recursive function
        // If the pivot is our item, then return it
        if (index == pivotindex)
            return pivotiter;
    
        // RECURSIVE CASE of former recursive function
        if (index < pivotindex)
        {   // Recurse on range of items preceding the pivot
            last = pivotiter;
        }
        else
        {   // Recurse on range of items following the pivot
            index -= pivotindex+1;
            first = pivotiter+1;
        }
        // Tail call is gone, replaced by loop
        // return quickselect(first, last, index);
    }
}


// trySmallData
// Call quickselect on a small dataset, print results.
void trySmallData()
{
    //vector<int> v { 5, 3, 2, 8, 5, 9, 10, 4, 6, 1, -4 };
    // No C++11 initializer lists in Visual Studio 2012 :-(
    const size_t size = 11;
    int data[size] = { 5, 3, 2, 8, 5, 9, 10, 4, 6, 1, -4 };
    vector<int> v(data, data+size);
    cout << "Quickselect on a vector of size " << size << endl;
    cout << endl;

    // Print unaltered data
    cout << "Before:" << endl;
    for (auto val : v)
        cout << val << " ";
    cout << endl;
    cout << endl;

    // Do the selection
    size_t index = 7;  // Index to select
    cout << "DOING SELECTION (index = " << index << ") ... ";
    vector<int>::iterator pos = quickselect(v.begin(), v.end(), index);
    cout << "DONE" << endl;
    cout << endl;

    // Print results
    cout << "After:" << endl;
    for (auto val : v)
        cout << val << " ";
    cout << endl;
    cout << endl;

    cout << "Value at returned location: " << *pos << endl;
    cout << endl;
}


// tryBigData
// Call quickselect on a dataset of specified size, print results.
// Generated data are "messy". Average-case behavior (more or less) is
// demonstrated.
void tryBigData(size_t size)
{
    vector<int> v(size);
    cout << "Quickselect on a vector of size " << size 
         << " (messy data)" << endl;
    cout << endl;

    // Create messy data
    const double phi = (1.+sqrt(5.))/2.;
    for (size_t i = 0; i < size; ++i)
    {
        double x = double(i+1)*phi;
        int n = int((x - int(x)) * 1000000000.);
        v[i] = n;
    }

    // Do the selection
    size_t index = size-1;  // Index to select
    cout << "DOING SELECTION (index = " << index << ") ... ";
    cout.flush();
    vector<int>::iterator pos = quickselect(v.begin(), v.end(), index);
    cout << "DONE" << endl;
    cout << endl;

    // Print results
    cout << "Value at returned location: " << *pos << endl;
    cout << endl;
}


// tryKillerData
// Call quickselect on a dataset of specified size, print results.
// Generated data form the "killer" sequence for our particular flavor
// of Quickselect. Worst-case behavior is demonstrated.
void tryKillerData(size_t size)
{
    vector<int> v(size);
    cout << "Quickselect on a vector of size " << size 
         << " (killer data)" << endl;
    cout << endl;

    // Create killer data
    for (size_t i = 0; i < size; ++i)
        v[i] = int(i);
    if (size >= 2)
    {
        size_t esize = size - (size % 2);  // Greatest even num <= size
        for (size_t i = esize-2; ; i -= 2)
        {
            size_t m = i+(esize-i)/2;
            swap(v[i+1], v[m]);

            if (i == 0)
                break;
        }
    }

    // Do the selection
    size_t index = size-1;  // Index to select
    cout << "DOING SELECTION (index = " << index << ") ... ";
    cout.flush();
    vector<int>::iterator pos = quickselect(v.begin(), v.end(), index);
    cout << "DONE" << endl;
    cout << endl;

    // Print results
    cout << "Value at returned location: " << *pos << endl;
    cout << endl;
}


// Main program
// Demonstrates function quickselect
int main()
{
    // TO CHANGE THE DATASET USED:
    // - Uncomment a different line below
    // - Change the parameter in the last 2 functions
    trySmallData();
    //tryBigData(100000);     // Parameter is size of dataset
    //tryKillerData(100000);  // Parameter is size of dataset

    cout << "*** To change dataset, edit function main in source ***"
         << endl;
    cout << endl;

    cout << "Press ENTER to quit ";
    while (cin.get() != '\n') ;

    return 0;
}

